"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from xanthus.models.neural import (
    GeneralizedMatrixFactorizationModel,
    NeuralMatrixFactorizationModel,
    MultiLayerPerceptronModel,
)
from xanthus.models.baseline import MatrixFactorizationModel, PopRankModel
